/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** nes6502.c
**
** NES custom 6502 (2A03) CPU implementation
*/

/* Revision history:
**
** 1.1   12/08/99
** - lots of fixes
** - irq and nmi were not sucking up any cycles like they should have
**
** 1.0
** - initial release
*/

#include "types.h"
#include "nes6502.h"

#ifdef NES6502_DEBUG
#include "log.h"
#endif

/* Treat invalid opcodes as NOPs (convention) */
#ifdef   X
#error   X already defined, use something else
#endif /* X */
#define  X  2

static const int cycle_table[256] =
{
/*       =============================================================== */
/*        0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F  */
/*       =============================================================== */
/* 0 */   7,  6,  X,  X,  X,  3,  5,  X,  3,  2,  2,  X,  X,  4,  6,  X,
/* 1 */   2,  5,  X,  X,  X,  4,  6,  X,  2,  4,  X,  X,  X,  4,  7,  X,
/* 2 */   6,  6,  X,  X,  3,  3,  5,  X,  4,  2,  2,  X,  4,  4,  6,  X,
/* 3 */   2,  5,  X,  X,  X,  4,  6,  X,  2,  4,  X,  X,  X,  4,  7,  X,
/* 4 */   6,  6,  X,  X,  X,  3,  5,  X,  3,  2,  2,  X,  3,  4,  6,  X,
/* 5 */   2,  5,  X,  X,  X,  4,  6,  X,  2,  4,  X,  X,  X,  4,  7,  X,
/* 6 */   6,  6,  X,  X,  X,  3,  5,  X,  4,  2,  2,  X,  5,  4,  6,  X,
/* 7 */   2,  5,  X,  X,  X,  4,  6,  X,  2,  4,  X,  X,  X,  4,  7,  X,
/* 8 */   X,  6,  X,  X,  3,  3,  3,  X,  2,  X,  2,  X,  4,  4,  4,  X,
/* 9 */   2,  6,  X,  X,  4,  4,  4,  X,  2,  5,  2,  X,  X,  5,  X,  X,
/* A */   2,  6,  2,  X,  3,  3,  3,  X,  2,  2,  2,  X,  4,  4,  4,  X,
/* B */   2,  5,  X,  X,  4,  4,  4,  X,  2,  4,  2,  X,  4,  4,  4,  X,
/* C */   2,  6,  X,  X,  3,  3,  5,  X,  2,  2,  2,  X,  4,  4,  6,  X,
/* D */   2,  5,  X,  X,  X,  4,  6,  X,  2,  4,  X,  X,  X,  4,  7,  X,
/* E */   2,  6,  X,  X,  3,  3,  5,  X,  2,  2,  2,  X,  4,  4,  6,  X,
/* F */   2,  5,  X,  X,  X,  4,  6,  X,  2,  4,  X,  X,  X,  4,  7,  X
};

#undef   X


/*
** Miscellaneous macros
*/
#define IRQ() \
{ \
   PUSH_PC(); \
   CLEAR_FLAG(B_FLAG); \
   PUSH_REG(P); \
   SET_FLAG(I_FLAG); \
   PC = READ_ADDRESS(IRQ_VECTOR); \
   irq_pending = FALSE; \
   instruction_cycles += INTERRUPT_CYCLES; \
}

#define NMI() \
{ \
   PUSH_PC(); \
   PUSH_REG(P); \
   CLEAR_FLAG(B_FLAG); \
   PC = READ_ADDRESS(NMI_VECTOR); \
   nmi_pending = FALSE; \
   instruction_cycles += INTERRUPT_CYCLES; \
}

#define CHECK_PAGE_CROSS(wOldPC, wNewPC) \
{ \
   if (((wOldPC) >> 8) != ((wNewPC) >> 8)) \
      instruction_cycles++; \
}

/*
** Addressing mode macros
**
** The DATA flavor fill in the bBucket that you pass in,
** the ADDR variety also fill in the wAddr
*/
#define IMMEDIATE_BYTE(bBucket) \
{ \
   bBucket = READ_OPCODE(PC); \
   PC++; \
}

#define ABSOLUTE_ADDR(wAddr) \
{ \
   wAddr = READ_ADDRESS(PC); \
   PC += 2; \
}

#define ABSOLUTE(wAddr, bBucket) \
{ \
   ABSOLUTE_ADDR(wAddr); \
   bBucket = MEM_READ(wAddr); \
}

#define ABSOLUTE_BYTE(bBucket) ABSOLUTE(wScratch, bBucket)


#define ABS_IND_X_ADDR(wAddr) \
{ \
   wAddr = READ_ADDRESS(PC) + X; \
   CHECK_PAGE_CROSS(PC, PC + 2); \
   PC += 2; \
}

#define ABS_IND_X(wAddr, bBucket) \
{ \
   ABS_IND_X_ADDR(wAddr); \
   bBucket = MEM_READ(wAddr); \
}

#define ABS_IND_X_BYTE(bBucket) ABS_IND_X(wScratch, bBucket)


#define ABS_IND_Y_ADDR(wAddr) \
{ \
   wAddr = READ_ADDRESS(PC) + Y; \
   CHECK_PAGE_CROSS(PC, PC + 2); \
   PC += 2; \
}

#define ABS_IND_Y(wAddr, bBucket) \
{ \
   ABS_IND_Y_ADDR(wAddr); \
   bBucket = MEM_READ(wAddr); \
}

#define ABS_IND_Y_BYTE(bBucket) ABS_IND_Y(wScratch, bBucket)


#define ZERO_PAGE_ADDR(bAddr) \
{ \
   bAddr = READ_OPCODE(PC); \
   PC++; \
}

#define ZERO_PAGE(bAddr, bBucket) \
{ \
   ZERO_PAGE_ADDR(bAddr); \
   bBucket = ZP_READ(bAddr); \
}

#define ZERO_PAGE_BYTE(bBucket) ZERO_PAGE(bScratch, bBucket)


/* Zero-page indexed Y doesn't really exist, just for LDX / STX */
#define ZP_IND_X_ADDR(bAddr) \
{ \
   bAddr = READ_OPCODE(PC) + X; \
   PC++; \
}

#define ZP_IND_X(bAddr, bBucket) \
{ \
   ZP_IND_X_ADDR(bAddr); \
   bBucket = ZP_READ(bAddr); \
}

#define ZP_IND_X_BYTE(bBucket) ZP_IND_X(bScratch, bBucket)


#define ZP_IND_Y_ADDR(bAddr) \
{ \
   bAddr = READ_OPCODE(PC) + Y; \
   PC++; \
}

#define ZP_IND_Y(bAddr, bBucket) \
{ \
   ZP_IND_Y_ADDR(bAddr); \
   bBucket = ZP_READ(bAddr); \
}

#define ZP_IND_Y_BYTE(bBucket) ZP_IND_Y(bScratch, bBucket)

/*
** For jump relative instructions
** (BCC, BCS, BEQ, BMI, BNE, BPL, BVC, BVS)
*/
#define RELATIVE_JUMP(bCondition) \
{ \
   if (bCondition) \
   /* jump! */ \
   { \
      wScratch = PC++; \
      bScratch = READ_OPCODE(wScratch); \
      /* Make offset signed, so jump will be back if negative */ \
      PC += ((int8) bScratch); \
      CHECK_PAGE_CROSS(wScratch, PC); \
      /* Jump taken, extra cycle penalty */ \
      instruction_cycles++; \
   } \
   else \
      PC++; \
}

#define  BUGGY_READ(wAddr) \
   (READ_OPCODE((wAddr)) | ((uint16) READ_OPCODE((uint16) ((wAddr) & ~0xFF)) << 8))

/* Pure indirect - for one lousy instruction! */
#define INDIRECT_ADDR(wAddr) \
{ \
   wScratch = READ_ADDRESS(PC); \
   PC += 2; \
   /* bug in crossing page boundaries */ \
   if (0xFF == (uint8) wScratch) \
      wAddr = BUGGY_READ(wScratch); \
   else \
      wAddr = READ_ADDRESS(wScratch); \
}

/*
** This is actually indexed indirect, but I call it
** indirect X to avoid confusion
*/
#define INDIR_X_ADDR(wAddr) \
{ \
   bScratch = READ_OPCODE(PC) + X; \
   PC++; \
   wAddr = ZP_ADDRESS(bScratch); \
}

#define INDIR_X_BYTE(bBucket) \
{ \
   bScratch = X + READ_OPCODE(PC); \
   PC++; \
   bBucket = MEM_READ(ZP_ADDRESS(bScratch)); \
}

/*
** This is actually indirect indexed, but I call it
** indirect y to avoid confusion
*/
#define INDIR_Y_ADDR(wAddr) \
{ \
   bScratch = READ_OPCODE(PC); \
   PC++; \
   wAddr = ZP_ADDRESS(bScratch) + Y; \
}

#define INDIR_Y_BYTE(bBucket) \
{ \
   bScratch = READ_OPCODE(PC); \
   PC++; \
   bBucket = MEM_READ(ZP_ADDRESS(bScratch) + Y); \
}


/*
** Instruction macros
*/

/* Warning! NES CPU has no decimal mode, so by default this does no BCD! */
#ifdef NES6502_DECIMAL
#define ADC(bVal) \
{ \
   if (P & D_FLAG) \
   { \
      wScratch = (A & 0x0F) + ((bVal) & 0x0F) + (P & C_FLAG); \
      if (wScratch > 0x09) \
         wScratch += 0x06; \
      if (wScratch <= 0x0F) \
         wScratch = (wScratch & 0x0F) + (A & 0xF0) + ((bVal) & 0xF0); \
      else \
         wScratch = (wScratch & 0x0F) + (A & 0xF0) + ((bVal) & 0xF0) + 0x10; \
      COND_SET_CLEAR(!((A + (bVal) + (P & C_FLAG)) & 0xff), Z_FLAG); \
      COND_SET_CLEAR(wScratch & 0x80, N_FLAG); \
      COND_SET_CLEAR(((~(A ^ (bVal))) & (A ^ wScratch) & 0x80), V_FLAG); \
      if ((wScratch & 0x1F0) > 0x90) \
         wScratch += 0x60; \
      COND_SET_CLEAR((wScratch & 0xFF0) > 0xF0, C_FLAG); \
      A = (uint8) (wScratch & 0xFF); \
   } \
   else \
   { \
      wScratch = A + (bVal) + (P & C_FLAG); \
      /* Set V on overflow */ \
      COND_SET_CLEAR(((~(A ^ (bVal))) & (A ^ wScratch) & 0x80), V_FLAG); \
      /* Set C on carry */ \
      COND_SET_CLEAR((wScratch & 0x100), C_FLAG); \
      A = (uint8) (wScratch & 0xFF); \
      SET_NZ_FLAGS(A); \
   }\
}
#else
#define ADC(bVal) \
{ \
   wScratch = A + (bVal) + (P & C_FLAG); \
   /* Set V on overflow */ \
   COND_SET_CLEAR(((~(A ^ (bVal))) & (A ^ wScratch) & 0x80), V_FLAG); \
   /* Set C on carry */ \
   COND_SET_CLEAR((wScratch & 0x100), C_FLAG); \
   A = (uint8) (wScratch & 0xFF); \
   SET_NZ_FLAGS(A); \
}
#endif /* NES6502_DECIMAL */

#define AND(bVal) \
{ \
   A &= (bVal); \
   SET_NZ_FLAGS(A); \
}

#define ASL(bVal) \
{ \
   COND_SET_CLEAR((bVal) & 0x80, C_FLAG); \
   bVal <<= 1; \
   SET_NZ_FLAGS(bVal);\
}

#define BCC() \
{ \
   RELATIVE_JUMP((IS_FLAG_CLEAR(C_FLAG))); \
}

#define BCS() \
{ \
   RELATIVE_JUMP((IS_FLAG_SET(C_FLAG))); \
}

#define BEQ() \
{ \
   RELATIVE_JUMP((IS_FLAG_SET(Z_FLAG))); \
}

#define BIT(bVal) \
{ \
   COND_SET_CLEAR(0 == ((bVal) & A), Z_FLAG);\
   CLEAR_FLAG(N_FLAG | V_FLAG); \
   /* move bit 7/6 of data into N/V flags */ \
   SET_FLAG(((bVal) & (N_FLAG | V_FLAG))); \
}

#define BMI() \
{ \
   RELATIVE_JUMP((IS_FLAG_SET(N_FLAG))); \
}

#define BNE() \
{ \
   RELATIVE_JUMP((IS_FLAG_CLEAR(Z_FLAG))); \
}

#define BPL() \
{ \
   RELATIVE_JUMP((IS_FLAG_CLEAR(N_FLAG))); \
}

/* Software interrupt type thang */
#define BRK() \
{ \
   PC++; \
   PUSH_PC(); \
   SET_FLAG(B_FLAG); \
   PUSH_REG(P); \
   SET_FLAG(I_FLAG); \
   PC = READ_ADDRESS(IRQ_VECTOR); \
}

#define BVC() \
{ \
   RELATIVE_JUMP((IS_FLAG_CLEAR(V_FLAG))); \
}

#define BVS() \
{ \
   RELATIVE_JUMP((IS_FLAG_SET(V_FLAG))); \
}

#define CLC() \
{ \
   CLEAR_FLAG(C_FLAG); \
}

#define CLD() \
{ \
   CLEAR_FLAG(D_FLAG); \
}

#define CLI() \
{ \
   CLEAR_FLAG(I_FLAG); \
   if (TRUE == irq_pending) \
      IRQ(); \
}

#define CLV() \
{ \
   CLEAR_FLAG(V_FLAG); \
}

#define CMP(bVal) \
{ \
   wScratch = A - (bVal); \
   /* C is clear when data > A, or it rolls around (0x8000-0xFFFF) */ \
   COND_SET_CLEAR(0 == (wScratch & 0x8000), C_FLAG); \
   SET_NZ_FLAGS(wScratch & 0xFF); /* handles Z flag */ \
}

#define CPX(bVal) \
{ \
   wScratch = X - (bVal); \
   /* C is clear when data > X, or it rolls around (0x8000-0xFFFF) */ \
   COND_SET_CLEAR(0 == (wScratch & 0x8000), C_FLAG); \
   SET_NZ_FLAGS(wScratch & 0xFF); /* handles Z flag */ \
}

#define CPY(bVal) \
{ \
   wScratch = Y - (bVal); \
   /* C is clear when data > Y, or it rolls around (0x8000-0xFFFF) */ \
   COND_SET_CLEAR(0 == (wScratch & 0x8000), C_FLAG); \
   SET_NZ_FLAGS(wScratch & 0xFF); /* handles Z flag */ \
}

#define DEC(bVal) \
{ \
   bVal--; \
   SET_NZ_FLAGS(bVal);\
}

#define DEX() \
{ \
   X--; \
   SET_NZ_FLAGS(X);\
}

#define DEY() \
{ \
   Y--; \
   SET_NZ_FLAGS(Y);\
}

#define EOR(bVal) \
{ \
   A ^= (bVal); \
   SET_NZ_FLAGS(A);\
}

#define INC(bVal) \
{ \
   bVal++; \
   SET_NZ_FLAGS(bVal);\
}

#define INX() \
{ \
   X++; \
   SET_NZ_FLAGS(X);\
}

#define INY() \
{ \
   Y++; \
   SET_NZ_FLAGS(Y);\
}

#define JMP(wAddr) \
{ \
   PC = (wAddr); \
}

#define JSR() \
{ \
   PC++; \
   PUSH_PC(); \
   PC--; \
   PC = READ_ADDRESS(PC); \
}

#define LDA(bVal) \
{ \
   A = (bVal); \
   SET_NZ_FLAGS(A);\
}

#define LDX(bVal) \
{ \
   X = (bVal); \
   SET_NZ_FLAGS(X);\
}

#define LDY(bVal) \
{ \
   Y = (bVal); \
   SET_NZ_FLAGS(Y);\
}

#define LSR(bVal) \
{ \
   COND_SET_CLEAR((bVal) & 0x01, C_FLAG); \
   bVal >>= 1; \
   SET_NZ_FLAGS(bVal);\
}

#define NOP() \
{ \
}

#define ORA(bVal) \
{ \
   A |= (bVal); \
   SET_NZ_FLAGS(A);\
}

#define PHA() \
{ \
   PUSH_REG(A); \
}

#define PHP() \
{ \
   /* B flag is pushed on stack as well */ \
   PUSH_REG((P | B_FLAG)); \
}

#define PLA() \
{ \
   PULL_REG(A); \
   SET_NZ_FLAGS(A); \
}

#define PLP() \
{ \
   PULL_REG(P); \
   SET_FLAG(R_FLAG); /* ensure reserved flag is set */ \
   /* 9/9/99 Is this possible?  Nested IRQ's? */ \
   if ((TRUE == irq_pending) && (IS_FLAG_CLEAR(I_FLAG))) \
      IRQ(); \
}

/* 9-bit rotation (carry flag used for rollover) */
#define ROL(bVal) \
{ \
   bScratch = P & C_FLAG; \
   COND_SET_CLEAR((bVal) & 0x80, C_FLAG); \
   bVal <<= 1; \
   bVal |= bScratch; \
   SET_NZ_FLAGS(bVal);\
}

#define ROR(bVal) \
{ \
   bScratch = P & C_FLAG; \
   COND_SET_CLEAR((bVal) & 0x01, C_FLAG); \
   bVal >>= 1; \
   if (bScratch) \
      bVal |= 0x80; \
   SET_NZ_FLAGS(bVal);\
}

#define RTI() \
{ \
   PULL_REG(P); \
   SET_FLAG(R_FLAG); /* ensure reserved flag is set */ \
   PULL_PC(); \
   /* Check for pending IRQs */ \
   if ((TRUE == irq_pending) && (IS_FLAG_CLEAR(I_FLAG))) \
      IRQ(); \
}

#define RTS() \
{ \
   PULL_PC(); \
   PC++; \
}

/* Warning! NES CPU has no decimal mode, so by default this does no BCD! */
#ifdef NES6502_DECIMAL
#define SBC(bVal) \
{ \
   /* NOT(C) is considered borrow */ \
   wScratch = A - (bVal) - ((P & C_FLAG) ^ C_FLAG); \
   if (P & D_FLAG) \
   { \
      uint16 wTemp; \
      wTemp = (A & 0x0F) - ((bVal) & 0x0F) - ((P & C_FLAG) ^ C_FLAG); \
      if (wTemp & 0x10) \
         wTemp = ((wTemp - 6) & 0x0F) | ((A & 0xF0) - ((bVal) & 0xF0) - 0x10); \
      else \
         wTemp = (wScratch & 0x0F) | ((A & 0xF0) - ((bVal) & 0xF0)); \
      if (wTemp & 0x100) \
         wTemp -= 0x60; \
      COND_SET_CLEAR(wScratch < 0x100, C_FLAG); \
      SET_NZ_FLAGS(wScratch & 0xFF); \
      COND_SET_CLEAR(((A ^ wScratch) & 0x80) && ((A ^ bVal) & 0x80), V_FLAG); \
      A = (uint8) (wTemp & 0xFF); \
   } \
   else \
   { \
      /* Overflow */ \
      COND_SET_CLEAR(((A ^ (bVal)) & (A ^ wScratch) & 0x80), V_FLAG); \
      /* Carry bit is inverted */ \
      COND_SET_CLEAR(0 == (wScratch & 0x100), C_FLAG); \
      A = (uint8) (wScratch & 0xFF); \
      SET_NZ_FLAGS(A); \
   } \
}
#else
#define SBC(bVal) \
{ \
   /* NOT(C) is considered borrow */ \
   wScratch = A - (bVal) - ((P & C_FLAG) ^ C_FLAG); \
   /* Overflow */ \
   COND_SET_CLEAR(((A ^ (bVal)) & (A ^ wScratch) & 0x80), V_FLAG); \
   /* Carry bit is inverted */ \
   COND_SET_CLEAR(0 == (wScratch & 0x100), C_FLAG); \
   A = (uint8) (wScratch & 0xFF); \
   SET_NZ_FLAGS(A); \
}
#endif /* NES6502_DECIMAL */

#define SEC() \
{ \
   SET_FLAG(C_FLAG); \
}

#define SED() \
{ \
   SET_FLAG(D_FLAG); \
}

#define SEI() \
{ \
   SET_FLAG(I_FLAG); \
}

/* No STx instructions (handled in opcode switch) */
#define TAX() \
{ \
   X = A; \
   SET_NZ_FLAGS(X);\
}

#define TAY() \
{ \
   Y = A; \
   SET_NZ_FLAGS(Y);\
}

#define TSX() \
{ \
   X = S; \
   SET_NZ_FLAGS(X);\
}

#define TXA() \
{ \
   A = X; \
   SET_NZ_FLAGS(A);\
}

#define TXS() \
{ \
   S = X; \
}

#define TYA() \
{ \
   A = Y; \
   SET_NZ_FLAGS(A); \
}


/*
** Stack and data fetching macros
*/

/* get a 16-bit address from zero-page */
#define  ZP_ADDRESS(bAddr)       (ZP_OPCODE((bAddr)) | \
                                 (uint16) (ZP_OPCODE((bAddr) + 1) << 8))

/* Set/clear/test bits in the P register */
#define  SET_FLAG(mask)          P |= (mask)
#define  CLEAR_FLAG(mask)        P &= ~(mask)
#define  IS_FLAG_SET(mask)       (P & (mask))
#define  IS_FLAG_CLEAR(mask)     (!IS_FLAG_SET((mask)))

/* If test is true, then set mask, else clear mask */
#define  COND_SET_CLEAR(test, mask) \
{ \
   if ((test)) \
      SET_FLAG((mask)); \
   else \
      CLEAR_FLAG((mask)); \
}

/* Sets the Z and N flags based on given data, taken from precomputed table */
#define  SET_NZ_FLAGS(bVal)      P &= ~(N_FLAG | Z_FLAG); \
                                 P |= flag_table[(bVal)];

/* Push/pull registers and program counter to stack */
#define  PUSH_REG(bVal)          zpmem[STACK_PAGE + S--] = (uint8) (bVal)

#define  PULL_REG(bVal)          bVal = zpmem[++S + STACK_PAGE]

#define  PUSH_PC()               PUSH_REG((uint8) (PC >> 8)); \
                                 PUSH_REG((uint8) (PC & 0xFF))

#define  PULL_PC()               PULL_REG(PC); \
                                 PC |= (zpmem[++S + STACK_PAGE] << 8);

/* Macros to read/write a byte of data */
#define  ZP_READ(bAddr)          zpmem[(bAddr)]
#define  ZP_WRITE(bAddr, bVal)   zpmem[(bAddr)] = (uint8) (bVal)
#define  ZP_OPCODE(bAddr)        ZP_READ(bAddr)

#define  MEM_READ(wAddr)         Read6502Memory((wAddr))
#define  MEM_WRITE(wAddr, bVal)  Write6502Memory((wAddr), (bVal))

#ifdef NES6502_FASTBANKS
#define  BANKED_MEMORY(wAddr)    nes6502_banks[(wAddr) >> 13][(wAddr)]
#else
#define  BANKED_MEMORY(wAddr)    nes6502_banks[(wAddr) >> 13][(wAddr) & 0x1FFF]
#endif /* NES6502_FASTBANKS */


/* static data */
static nes6502_memread *pmem_read, *pmr;
static nes6502_memwrite *pmem_write, *pmw;
static uint8 flag_table[256];
static uint32 reg_PC;
static uint8 reg_A, reg_P, reg_X, reg_Y, reg_S;
static boolean irq_pending, nmi_pending;

#ifdef NES6502_DEBUG
static uint32 trap_addr;
static boolean trace;
#endif /* NES6502_DEBUG */

/* Keep track of how many cycles have elapsed */
static int total_cycles = 0;
static uint8 *zpmem = 0;
static uint8 *nes6502_banks[8]; /* Global 8kB banks */

/* Fast opcode reading routine */
#ifdef WIN32
static uint8 READ_OPCODE(uint32 address)
#else
static __inline__ uint8 READ_OPCODE(uint32 address)
#endif /* WIN32 */
{
#ifdef NES6502_DEBUG
   if (address >= 0x800 && address < 0x6000)
   {
      char error[128];

      sprintf(error, "FATAL: opcode fetch from $%04X: unpredictable results\n",
              address);
      ASSERT_MSG(error);
   }
#endif /* NES6502_DEBUG */

   return BANKED_MEMORY(address);
}

#ifdef NES6502_LITTLE_ENDIAN
#ifdef NES6502_FASTBANKS
#define  READ_ADDRESS(wAddr)   (uint16)(*(uint16 *) (nes6502_banks[(wAddr) >> 13] + (wAddr)))
#else
#define  READ_ADDRESS(wAddr)   (uint16)(*(uint16 *) (nes6502_banks[(wAddr) >> 13] + ((wAddr) & 0x1FFF)))
#endif /* NES6502_FASTBANKS */
#else
static inline uint16 READ_ADDRESS(uint32 address)
{
#ifdef NES6502_FASTBANKS
#if TARGET_CPU_PPC
        return __lhbrx(NES6502_Banks[address >> 13], address);
#else
        uint16 x = *(uint16 *)(nes6502_banks[address >> 13] + address);
        return (x << 8) | (x >> 8);
#endif /* TARGET_CPU_PPC */
#else
#if TARGET_CPU_PPC
        return __lhbrx(nes6502_banks[address >> 13], address & 0x1FFF);
#else
        uint16 x = *(uint16 *)(nes6502_banks[address >> 13] + (address & 0x1FFF));
        return (x << 8) | (x >> 8);
#endif /* TARGET_CPU_PPC */
#endif /* NES6502_FASTBANKS */
}
#endif /* NES6502_LITTLE_ENDIAN */


/* read a byte of 6502 memory */
static uint8 Read6502Memory(uint32 address)
{
   /* Zero page RAM */
   if (address < 0x800)
      return zpmem[address];
   else if (address >= 0x8000)
      return BANKED_MEMORY(address);
   else
   {
      pmr = pmem_read;

      while (pmr->min_range != 0xFFFFFFFF)
      {
         if ((address >= pmr->min_range) && (address <= pmr->max_range))
            return pmr->read_func(address);

         pmr++;
      }
   }

   return BANKED_MEMORY(address);
}

/* write a byte of data to 6502 memory */
static void Write6502Memory(uint32 address, uint8 value)
{
   if (address < 0x800)
   {
      zpmem[address] = value;
      return;
   }
   else
   {
      pmw = pmem_write;

      while (pmw->min_range != 0xFFFFFFFF)
      {
         if ((address >= pmw->min_range) && (address <= pmw->max_range))
         {
            pmw->write_func(address, value);
            return;
         }
         pmw++;
      }
   }

   BANKED_MEMORY(address) = value;
}

/* set the current context */
void nes6502_setcontext(nes6502_context *cpu)
{
   int loop;

   /* Set the page pointers */
   for (loop = 0; loop < 8; loop++)
      nes6502_banks[loop] = cpu->mem_page[loop];

   zpmem = nes6502_banks[0];  /* quicker zero-page references */

   pmem_read = cpu->read_handler;
   pmem_write = cpu->write_handler;

   reg_PC = cpu->pc_reg;
   reg_A = cpu->a_reg;
   reg_P = cpu->p_reg;
   reg_X = cpu->x_reg;
   reg_Y = cpu->y_reg;
   reg_S = cpu->s_reg;
   irq_pending = cpu->irq_pending;
   nmi_pending = cpu->nmi_pending;

#ifdef NES6502_DEBUG
   trap_addr = cpu->trap_addr;
   trace = cpu->trace;
#endif /* NES6502_DEBUG */
}

/* get the current context */
void nes6502_getcontext(nes6502_context *cpu)
{
   int loop;

   /* Set the page pointers */
   for (loop = 0; loop < 8; loop++)
      cpu->mem_page[loop] = nes6502_banks[loop];

   cpu->read_handler = pmem_read;
   cpu->write_handler = pmem_write;

   cpu->pc_reg = reg_PC;
   cpu->a_reg = reg_A;
   cpu->p_reg = reg_P;
   cpu->x_reg = reg_X;
   cpu->y_reg = reg_Y;
   cpu->s_reg = reg_S;
   cpu->irq_pending = irq_pending;
   cpu->nmi_pending = nmi_pending;

#ifdef NES6502_DEBUG
   cpu->trap_addr = trap_addr;
   cpu->trace = trace;
#endif
}

/* DMA a byte of data from ROM */
uint8 nes6502_dma(uint32 address)
{
   return BANKED_MEMORY(address);
}

/* get number of elapsed cycles */
int nes6502_getcycles(boolean reset_flag)
{
   int cycles = total_cycles;

   if (reset_flag)
      total_cycles = 0;

   return cycles;
}

/* Execute instructions until count expires
**
** Returns the number of cycles *actually* executed
** (note that this can be from 0-6 cycles more than you wanted)
*/
int nes6502_execute(int remaining_cycles)
{
   int elapsed_cycles = 0, instruction_cycles;
   uint8 opcode;

   /* Temporary/address/data variables for the macros */
   uint8 bScratch, bLoc, bData;
   uint32 wScratch, wLoc;

   uint32 PC = reg_PC;
   uint8 A = reg_A, X = reg_X, Y = reg_Y, P = reg_P, S = reg_S;

   /* Continue until we run out of cycles */
   while (remaining_cycles > 0)
   {
#ifdef NES6502_DEBUG
      /* turn tracing on when reached trap address */
/*
      if (PC == trap_addr)
         trace = TRUE;
      if (trace)
      {
         if (0 == dbg_nes6502())
            return PC;
      }
*/
#endif /* NES6502_DEBUG */

      instruction_cycles = 0;

      /* NMI has highest priority */
      if (nmi_pending)
         NMI();
      /* IRQ has lowest priority */
      if (irq_pending && IS_FLAG_CLEAR(I_FLAG))
         IRQ();

      /* Fetch instruction, calculate cycle time */
      opcode = READ_OPCODE(PC);
      PC++;
      instruction_cycles += cycle_table[opcode];

      /* Execute instruction */
      switch (opcode)
      {
      /* kill switch for CPU emulation */
      case 0xFF:
         goto _done;
         break;
         
      /* ADC */
      case 0x6D:  ABSOLUTE_BYTE(bData);  ADC(bData); break;
      case 0x65:  ZERO_PAGE_BYTE(bData); ADC(bData); break;
      case 0x69:  IMMEDIATE_BYTE(bData); ADC(bData); break;
      case 0x7D:  ABS_IND_X_BYTE(bData); ADC(bData); break;
      case 0x79:  ABS_IND_Y_BYTE(bData); ADC(bData); break;
      case 0x61:  INDIR_X_BYTE(bData);   ADC(bData); break;
      case 0x71:  INDIR_Y_BYTE(bData);   ADC(bData); break;
      case 0x75:  ZP_IND_X_BYTE(bData);  ADC(bData); break;

      /* AND */
      case 0x2D:  ABSOLUTE_BYTE(bData);  AND(bData); break;
      case 0x25:  ZERO_PAGE_BYTE(bData); AND(bData); break;
      case 0x29:  IMMEDIATE_BYTE(bData); AND(bData); break;
      case 0x3D:  ABS_IND_X_BYTE(bData); AND(bData); break;
      case 0x39:  ABS_IND_Y_BYTE(bData); AND(bData); break;
      case 0x21:  INDIR_X_BYTE(bData);   AND(bData); break;
      case 0x31:  INDIR_Y_BYTE(bData);   AND(bData); break;
      case 0x35:  ZP_IND_X_BYTE(bData);  AND(bData); break;

      /* ASL */
      case 0x0A:  ASL(A); break;
      case 0x0E:  ABSOLUTE(wLoc, bData);  ASL(bData); MEM_WRITE(wLoc, bData); break;
      case 0x06:  ZERO_PAGE(bLoc, bData); ASL(bData); ZP_WRITE(bLoc, bData); break;
      case 0x1E:  ABS_IND_X(wLoc, bData); ASL(bData); MEM_WRITE(wLoc, bData); break;
      case 0x16:  ZP_IND_X(bLoc, bData);  ASL(bData); ZP_WRITE(bLoc, bData); break;

      /* Relative branches */
      case 0x90:  BCC(); break;
      case 0xB0:  BCS(); break;
      case 0xF0:  BEQ(); break;
      case 0x30:  BMI(); break;
      case 0xD0:  BNE(); break;
      case 0x10:  BPL(); break;
      case 0x50:  BVC(); break;
      case 0x70:  BVS(); break;

      /* BIT */
      case 0x2C:  ABSOLUTE_BYTE(bData);  BIT(bData); break;
      case 0x24:  ZERO_PAGE_BYTE(bData); BIT(bData); break;

      /* BRK */
      case 0x00:  BRK(); break;

      /* Clear flag bits */
      case 0x18:  CLC(); break;
      case 0xD8:  CLD(); break;
      case 0x58:  CLI(); break;
      case 0xB8:  CLV(); break;

      /* CMP */
      case 0xCD:  ABSOLUTE_BYTE(bData);  CMP(bData); break;
      case 0xC5:  ZERO_PAGE_BYTE(bData); CMP(bData); break;
      case 0xC9:  IMMEDIATE_BYTE(bData); CMP(bData); break;
      case 0xDD:  ABS_IND_X_BYTE(bData); CMP(bData); break;
      case 0xD9:  ABS_IND_Y_BYTE(bData); CMP(bData); break;
      case 0xC1:  INDIR_X_BYTE(bData);   CMP(bData); break;
      case 0xD1:  INDIR_Y_BYTE(bData);   CMP(bData); break;
      case 0xD5:  ZP_IND_X_BYTE(bData);  CMP(bData); break;

      /* CPX */
      case 0xEC:  ABSOLUTE_BYTE(bData);  CPX(bData); break;
      case 0xE4:  ZERO_PAGE_BYTE(bData); CPX(bData); break;
      case 0xE0:  IMMEDIATE_BYTE(bData); CPX(bData); break;

      /* CPY */
      case 0xCC:  ABSOLUTE_BYTE(bData);  CPY(bData); break;
      case 0xC4:  ZERO_PAGE_BYTE(bData); CPY(bData); break;
      case 0xC0:  IMMEDIATE_BYTE(bData); CPY(bData); break;

      /* DEC */
      case 0xCE:  ABSOLUTE(wLoc, bData);  DEC(bData); MEM_WRITE(wLoc, bData); break;
      case 0xC6:  ZERO_PAGE(bLoc, bData); DEC(bData); ZP_WRITE(bLoc, bData); break;
      case 0xDE:  ABS_IND_X(wLoc, bData); DEC(bData); MEM_WRITE(wLoc, bData); break;
      case 0xD6:  ZP_IND_X(bLoc, bData);  DEC(bData); ZP_WRITE(bLoc, bData); break;

      /* DEX/Y */
      case 0xCA:  DEX(); break;
      case 0x88:  DEY(); break;

      /* EOR */
      case 0x4D:  ABSOLUTE_BYTE(bData);  EOR(bData); break;
      case 0x45:  ZERO_PAGE_BYTE(bData); EOR(bData); break;
      case 0x49:  IMMEDIATE_BYTE(bData); EOR(bData); break;
      case 0x5D:  ABS_IND_X_BYTE(bData); EOR(bData); break;
      case 0x59:  ABS_IND_Y_BYTE(bData); EOR(bData); break;
      case 0x41:  INDIR_X_BYTE(bData);   EOR(bData); break;
      case 0x51:  INDIR_Y_BYTE(bData);   EOR(bData); break;
      case 0x55:  ZP_IND_X_BYTE(bData);  EOR(bData); break;

      /* INC */
      case 0xEE:  ABSOLUTE(wLoc, bData);  INC(bData); MEM_WRITE(wLoc, bData); break;
      case 0xE6:  ZERO_PAGE(bLoc, bData); INC(bData); ZP_WRITE(bLoc, bData); break;
      case 0xFE:  ABS_IND_X(wLoc, bData); INC(bData); MEM_WRITE(wLoc, bData); break;
      case 0xF6:  ZP_IND_X(bLoc, bData);  INC(bData); ZP_WRITE(bLoc, bData); break;

      /* INX/Y */
      case 0xE8:  INX(); break;
      case 0xC8:  INY(); break;

      /* JMP/JSR */
      case 0x4C:  ABSOLUTE_ADDR(wLoc); JMP(wLoc); break;
      case 0x6C:  INDIRECT_ADDR(wLoc); JMP(wLoc); break;
      case 0x20:  JSR(); break;

      /* LDA */
      case 0xAD:  ABSOLUTE_BYTE(bData);  LDA(bData); break;
      case 0xA5:  ZERO_PAGE_BYTE(bData); LDA(bData); break;
      case 0xA9:  IMMEDIATE_BYTE(bData); LDA(bData); break;
      case 0xBD:  ABS_IND_X_BYTE(bData); LDA(bData); break;
      case 0xB9:  ABS_IND_Y_BYTE(bData); LDA(bData); break;
      case 0xA1:  INDIR_X_BYTE(bData);   LDA(bData); break;
      case 0xB1:  INDIR_Y_BYTE(bData);   LDA(bData); break;
      case 0xB5:  ZP_IND_X_BYTE(bData);  LDA(bData); break;

      /* LDX */
      case 0xAE:  ABSOLUTE_BYTE(bData);  LDX(bData); break;
      case 0xA6:  ZERO_PAGE_BYTE(bData); LDX(bData); break;
      case 0xA2:  IMMEDIATE_BYTE(bData); LDX(bData); break;
      case 0xBE:  ABS_IND_Y_BYTE(bData); LDX(bData); break;
      case 0xB6:  ZP_IND_Y_BYTE(bData);  LDX(bData); break;

      /* LDY */
      case 0xAC:  ABSOLUTE_BYTE(bData);  LDY(bData); break;
      case 0xA4:  ZERO_PAGE_BYTE(bData); LDY(bData); break;
      case 0xA0:  IMMEDIATE_BYTE(bData); LDY(bData); break;
      case 0xBC:  ABS_IND_X_BYTE(bData); LDY(bData); break;
      case 0xB4:  ZP_IND_X_BYTE(bData);  LDY(bData); break;

      /* LSR */
      case 0x4A:  LSR(A); break;
      case 0x4E:  ABSOLUTE(wLoc, bData);  LSR(bData); MEM_WRITE(wLoc, bData); break;
      case 0x46:  ZERO_PAGE(bLoc, bData); LSR(bData); ZP_WRITE(bLoc, bData); break;
      case 0x5E:  ABS_IND_X(wLoc, bData); LSR(bData); MEM_WRITE(wLoc, bData); break;
      case 0x56:  ZP_IND_X(bLoc, bData);  LSR(bData); ZP_WRITE(bLoc, bData); break;

      /* NOP */
      case 0xEA:  NOP(); break;

      /* ORA */
      case 0x0D:  ABSOLUTE_BYTE(bData);  ORA(bData); break;
      case 0x05:  ZERO_PAGE_BYTE(bData); ORA(bData); break;
      case 0x09:  IMMEDIATE_BYTE(bData); ORA(bData); break;
      case 0x1D:  ABS_IND_X_BYTE(bData); ORA(bData); break;
      case 0x19:  ABS_IND_Y_BYTE(bData); ORA(bData); break;
      case 0x01:  INDIR_X_BYTE(bData);   ORA(bData); break;
      case 0x11:  INDIR_Y_BYTE(bData);   ORA(bData); break;
      case 0x15:  ZP_IND_X_BYTE(bData);  ORA(bData); break;

      /* Push / pull regs off stack */
      case 0x48:  PHA(); break;
      case 0x08:  PHP(); break;
      case 0x68:  PLA(); break;
      case 0x28:  PLP(); break;

      /* ROL */
      case 0x2A:  ROL(A); break;
      case 0x2E:  ABSOLUTE(wLoc, bData);  ROL(bData); MEM_WRITE(wLoc, bData); break;
      case 0x26:  ZERO_PAGE(bLoc, bData); ROL(bData); ZP_WRITE(bLoc, bData); break;
      case 0x3E:  ABS_IND_X(wLoc, bData); ROL(bData); MEM_WRITE(wLoc, bData); break;
      case 0x36:  ZP_IND_X(bLoc, bData);  ROL(bData); ZP_WRITE(bLoc, bData); break;

      /* ROR */
      case 0x6A:  ROR(A); break;
      case 0x6E:  ABSOLUTE(wLoc, bData);  ROR(bData); MEM_WRITE(wLoc, bData); break;
      case 0x66:  ZERO_PAGE(bLoc, bData); ROR(bData); ZP_WRITE(bLoc, bData); break;
      case 0x7E:  ABS_IND_X(wLoc, bData); ROR(bData); MEM_WRITE(wLoc, bData); break;
      case 0x76:  ZP_IND_X(bLoc, bData);  ROR(bData); ZP_WRITE(bLoc, bData); break;

      /* RTI/S */
      case 0x40:  RTI(); break;
      case 0x60:  RTS(); break;

      /* SBC */
      case 0xED:  ABSOLUTE_BYTE(bData);  SBC(bData); break;
      case 0xE5:  ZERO_PAGE_BYTE(bData); SBC(bData); break;
      case 0xE9:  IMMEDIATE_BYTE(bData); SBC(bData); break;
      case 0xFD:  ABS_IND_X_BYTE(bData); SBC(bData); break;
      case 0xF9:  ABS_IND_Y_BYTE(bData); SBC(bData); break;
      case 0xE1:  INDIR_X_BYTE(bData);   SBC(bData); break;
      case 0xF1:  INDIR_Y_BYTE(bData);   SBC(bData); break;
      case 0xF5:  ZP_IND_X_BYTE(bData);  SBC(bData); break;

      /* Set flag reg bits */
      case 0x38:  SEC(); break;
      case 0xF8:  SED(); break;
      case 0x78:  SEI(); break;

      /* STA */
      case 0x8D:  ABSOLUTE_ADDR(wLoc);  MEM_WRITE(wLoc, A); break;
      case 0x85:  ZERO_PAGE_ADDR(bLoc); ZP_WRITE(bLoc, A); break;
      case 0x9D:  ABS_IND_X_ADDR(wLoc); MEM_WRITE(wLoc, A); break;
      case 0x99:  ABS_IND_Y_ADDR(wLoc); MEM_WRITE(wLoc, A); break;
      case 0x81:  INDIR_X_ADDR(wLoc);   MEM_WRITE(wLoc, A); break;
      case 0x91:  INDIR_Y_ADDR(wLoc);   MEM_WRITE(wLoc, A); break;
      case 0x95:  ZP_IND_X_ADDR(bLoc);  ZP_WRITE(bLoc, A); break;

      /* STX */
      case 0x8E:  ABSOLUTE_ADDR(wLoc);  MEM_WRITE(wLoc, X); break;
      case 0x86:  ZERO_PAGE_ADDR(bLoc); ZP_WRITE(bLoc, X); break;
      case 0x96:  ZP_IND_Y_ADDR(bLoc);  ZP_WRITE(bLoc, X); break;

      /* STY */
      case 0x8C:  ABSOLUTE_ADDR(wLoc);  MEM_WRITE(wLoc, Y); break;
      case 0x84:  ZERO_PAGE_ADDR(bLoc); ZP_WRITE(bLoc, Y); break;
      case 0x94:  ZP_IND_X_ADDR(bLoc);  ZP_WRITE(bLoc, Y); break;

      /* Inter-register transfer */
      case 0xAA:  TAX(); break;
      case 0xA8:  TAY(); break;
      case 0xBA:  TSX(); break;
      case 0x8A:  TXA(); break;
      case 0x9A:  TXS(); break;
      case 0x98:  TYA(); break;

      /* Bad instruction, most likely a bad ROM dump */
      /* Luckily, for the NES, we don't have any undocumented opcodes */
      default:
#ifdef NES6502_DEBUG
         {
            char error[128];
            sprintf(error, "FATAL: invalid opcode $%02X\n", opcode);
            ASSERT_MSG(error);
         }
#endif
         break;
      }

      /* Calculate remaining/elapsed clock cycles */
      remaining_cycles -= instruction_cycles;
      elapsed_cycles += instruction_cycles;
      total_cycles += instruction_cycles;
   }

_done:
   /* restore local copy of regs */
   reg_A = A;
   reg_X = X;
   reg_Y = Y;
   reg_P = P;
   reg_S = S;
   reg_PC = PC;

   /* Return our actual amount of executed cycles */
   return elapsed_cycles;
}

/* Initialize some CPU stuff */
void nes6502_init(void)
{
   int index;

   /* Build the N / Z flag lookup table */
   flag_table[0] = Z_FLAG;
   for (index = 1; index < 256; index++)
      flag_table[index] = (index & 0x80) ? N_FLAG : 0;
}

/* Issue a CPU Reset */
void nes6502_reset(void)
{
   reg_A = 0;
   reg_X = 0;
   reg_Y = 0;
   reg_P = Z_FLAG | R_FLAG;              /* Reserved bit always 1   */
   reg_S = 0xFF;                         /* Stack grows down $01FF->*/
   irq_pending = FALSE;                  /* No IRQs pending         */
   nmi_pending = FALSE;
   reg_PC = READ_ADDRESS(RESET_VECTOR);  /* Fetch reset vector      */
}

/* Non-maskable interrupt */
void nes6502_nmi(void)
{
   nmi_pending = TRUE;
}

/* Interrupt request */
void nes6502_irq(void)
{
   irq_pending = TRUE;
}
