/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** nes_apu.c
**
** NES APU emulation
*/

#include "types.h"
#include "cpu.h"
#include "nsf_play.h"
#include "nes_apu.h"
#include "log.h"

static square_t square[2];
static triangle_t triangle;
static noise_t noise;
static dpcm_t dpcm;
static uint8 apu_enable;

static boolean mix_enable[6];
static uint8 noise_lut[NOISE_LONG];

static apuqueue_t apu_q[APUQUEUE_SIZE];
static int q_head = 0;
static int q_tail = 0;

static apusound_t apu_ext = NULL;
static int apu_buf_size;
static float apu_incsize;

static int apu_dsp = APU_DSP_NONE;

static uint8 *pcm_data = NULL;
static int pcm_length = 0;

/* various sound constants for sound emulation */
/* vblank length table used for squares, triangle, noise */
static const uint8 vbl_length[32] =
{
   5, 127, 10, 1, 19,  2, 40,  3, 80,  4, 30,  5, 7,  6, 13,  7,
   6,   8, 12, 9, 24, 10, 48, 11, 96, 12, 36, 13, 8, 14, 16, 15
};

/* frequency limit of square channels */
static const int freq_limit[8] =
{
   0x3FF, 0x555, 0x666, 0x71C, 0x787, 0x7C1, 0x7E0, 0x7F0,
};

/* table of noise frequencies */
static const int noise_freq[16] =
{
   4, 8, 16, 32, 64, 96, 128, 160, 202, 254, 380, 508, 762, 1016, 2034, 2046
};

/* dpcm transfer freqs */
const int dpcm_clocks[16] =
{
   428, 380, 340, 320, 286, 254, 226, 214, 190, 160, 142, 128, 106, 85, 72, 54
};

/* ratios of pos/neg pulse for square waves */
/* 2/16 = 12.5%, 4/16 = 25%, 8/16 = 50%, 12/16 = 75% */
static const int duty_lut[4] =
{
   2, 4, 8, 12
};


static boolean apu_isqempty(void)
{
   return (q_head == q_tail);
}

static void apu_enqueue(apuqueue_t q)
{
   apu_q[q_head] = q;

   if (++q_head >= APUQUEUE_SIZE)
      q_head = 0;

   if (TRUE == apu_isqempty())
      log_printf("apu: queue overflow\n");
}

static apuqueue_t apu_dequeue(void)
{
   int loc;

   if (TRUE == apu_isqempty())
      log_printf("apu: queue empty\n");

   loc = q_tail++;
   if (q_tail >= APUQUEUE_SIZE)
      q_tail = 0;

   return apu_q[loc];
}

void apu_setchan(int chan, boolean enabled)
{
   mix_enable[chan] = enabled;
}


/* TODO: sound channels should *ALL* have decay, booster */

static int8 apu_square(square_t *chan)
{
   int env_delay;
   int sweep_delay;
   int8 output;

   /* reg0: 0-3=volume, 4=envelope, 5=hold, 6-7=duty cycle
   ** reg1: 0-2=sweep shifts, 3=sweep inc/dec, 4-6=sweep length, 7=sweep on
   ** reg2: 8 bits of freq
   ** reg3: 0-2=high freq, 7-4=vbl length counter
   */

   if (FALSE == chan->enabled)
      return 0;

   /* enveloping */
   env_delay = ((chan->regs[0] & 0x0F) + 1) * apu_buf_size;

   /* decay is at a rate of (env_regs + 1) / 240 secs */
   chan->env_phase -= 4;
   while (chan->env_phase < 0)
   {
      chan->env_phase += env_delay;
      if (chan->regs[0] & 0x20)
         chan->env_vol = (chan->env_vol + 1) & 15;
      else if (chan->env_vol < 15)
         chan->env_vol++;
   }

   /* vbl length counter */
   if (chan->vbl_length > 0 && 0 == (chan->regs [0] & 0x20))
      chan->vbl_length--;

   if (0 == chan->vbl_length)
      return 0;

   /* freqsweeps */
   if ((chan->regs[1] & 0x80) && (chan->regs[1] & 7))
   {
      sweep_delay = (((chan->regs[1] >> 4) & 7) + 1) * apu_buf_size;
      chan->sweep_phase -= 2;
      while (chan->sweep_phase < 0)
      {
         chan->sweep_phase += sweep_delay;
         if (chan->regs[1] & 8)
            chan->freq -= chan->freq >> (chan->regs[1] & 7);
         else
            chan->freq += chan->freq >> (chan->regs[1] & 7);
      }
   }

   if ((0 == (chan->regs[1] & 8) && (chan->freq >> 16) > freq_limit[chan->regs[1] & 7])
       || (chan->freq >> 16) < 4)
      return 0;

   chan->phaseacc -= (float) apu_incsize; /* # of cycles per sample */

   while (chan->phaseacc < 0)
   {
      chan->phaseacc += (chan->freq >> 16);
      chan->adder = (chan->adder + 1) & 0x0F;
   }

   if (chan->regs[0] & 0x10) /* fixed volume */
      output = chan->regs[0] & 0x0F;
   else
      output = 15 - chan->env_vol;

   if (chan->adder < (duty_lut[chan->regs[0] >> 6]))
      output = -output;

   return output;
}

static int8 apu_triangle(triangle_t *chan)
{
   int freq;
   int8 output;
   /* reg0: 7=holdnote, 6-0=linear length counter
   ** reg2: low 8 bits of frequency
   ** reg3: 7-3=length counter, 2-0=high 3 bits of frequency
   */

   if (FALSE == chan->enabled)
      return 0;

   if (FALSE == chan->counter_started && 0 == (chan->regs[0] & 0x80))
   {
      if (chan->write_latency)
         chan->write_latency--;
      if (0 == chan->write_latency)
         chan->counter_started = TRUE;
   }

   if (chan->counter_started)
   {
      if (chan->linear_length > 0)
         chan->linear_length--;
      if (chan->vbl_length && 0 == (chan->regs[0] & 0x80))
            chan->vbl_length--;

      if (0 == chan->vbl_length)
         return 0;
   }

   if (0 == chan->linear_length)
      return 0;

   freq = (((chan->regs[3] & 7) << 8) + chan->regs[2]) + 1;

   if (freq < 4) /* inaudible */
      return 0;

   chan->phaseacc -= (float) apu_incsize; /* # of cycles per sample */
   while (chan->phaseacc < 0)
   {
      chan->phaseacc += freq;
      chan->adder = (chan->adder + 1) & 0x1F;
   
      output = (chan->adder & 7) << 1;
      if (chan->adder & 8)
         output = 0x10 - output;
      if (chan->adder & 0x10)
         output = -output;

      chan->output_vol = output;   
   }

   return (uint8) chan->output_vol;
}

static int8 apu_noise(noise_t *chan)
{
   int freq, env_delay;
   uint8 outvol;
   uint8 output;

   /* reg0: 0-3=volume, 4=envelope, 5=hold
   ** reg2: 7=small(93 byte) sample,3-0=freq lookup
   ** reg3: 7-4=vbl length counter
   */

   if (FALSE == chan->enabled)
      return 0;

   /* enveloping */
   env_delay = ((chan->regs[0] & 0x0F) + 1) * apu_buf_size;

   /* decay is at a rate of (env_regs + 1) / 240 secs */
   chan->env_phase -= 4;
   while (chan->env_phase < 0)
   {
      chan->env_phase += env_delay;
      if (chan->regs[0] & 0x20)
         chan->env_vol = (chan->env_vol + 1) & 15;
      else if (chan->env_vol < 15)
         chan->env_vol++;
   }

   /* length counter */
   if (0 == (chan->regs[0] & 0x20))
   {
      if (chan->vbl_length > 0)
         chan->vbl_length--;
   }

   if (0 == chan->vbl_length)
      return 0;

   freq = noise_freq[chan->regs[2] & 0x0F];
   chan->phaseacc -= (float) apu_incsize; /* # of cycles per sample */
   while (chan->phaseacc < 0)
   {
      chan->phaseacc += freq;

      chan->cur_pos++;
      if (93 == chan->cur_pos && (chan->regs[2] & 0x80))
         chan->cur_pos = 0;
      else if (NOISE_LONG == chan->cur_pos)
         chan->cur_pos = 0;
   }

   if (chan->regs[0] & 0x10) /* fixed volume */
      outvol = chan->regs[0] & 0x0F;
   else
      outvol = 15 - chan->env_vol;

   output = noise_lut[chan->cur_pos];
   if (output > outvol)
      output = outvol;

   if (noise_lut[chan->cur_pos] & 0x80) /* make it negative */
   //   output = -output;
   //return output;
      return (int8) -output;
   else
      return (int8) output;
}

static void apu_dpcmreset(dpcm_t *chan)
{
   chan->address = 0xC000 + (uint16) (chan->regs[2] << 6);
   chan->length = (uint16) (chan->regs[3] << 4) + 1;
   chan->bits_left = chan->length << 3;
   chan->irq_occurred = FALSE; 
}

/* TODO: centerline naughtiness */
static int8 apu_dpcm(dpcm_t *chan)
{
   int freq, bit_pos;

   /* reg0: 7=irq gen, 6=looping, 3-0=pointer to clock table
   ** reg1: output dc level, 6 bits unsigned
   ** reg2: 8 bits of 64-byte aligned address offset : $C000 + (value * 64)
   ** reg3: length, (value * 16) + 1
   */

   /* TODO: Ack! Ack! Phhbbtt! */
   if (FALSE == chan->enabled)
      goto _done;

   freq = dpcm_clocks[chan->regs[0] & 0x0F];
   chan->phaseacc -= (float) apu_incsize; /* # of cycles per sample */

   while (chan->phaseacc < 0)
   {
      chan->phaseacc += freq;

      if (0 == chan->length)
      {
         if (chan->regs[0] & 0x40)
            apu_dpcmreset(chan);
         else
         {
            if (chan->regs[0] & 0x80) /* irq gen */
            {
               chan->irq_occurred = TRUE;
               cpu_irq();
            }
            break;
         }
      }

      chan->bits_left--;
      bit_pos = 7 - (chan->bits_left & 7);
      if (7 == bit_pos)
      {
         chan->cur_byte = cpu_dma(chan->address);
         chan->address++;
         chan->length--;
      }

      if (chan->cur_byte & (1 << bit_pos))
         chan->regs[1]++;
      else
         chan->regs[1]--;
   }

_done:

   if ((int8) chan->regs[1] > 31)
      chan->regs[1] = 31;
   if ((int8) chan->regs[1] < -32)
      chan->regs[1] = -32;

   return chan->regs[1];
}

static void apu_regwrite(uint32 address, uint8 value)
{  
   int chan;

   switch (address)
   {
   /* squares */
   case APU_WRA0:
   case APU_WRB0:
      chan = (address & 4) ? 1 : 0;
      square[chan].regs[0] = value;
      break;

   case APU_WRA1:
   case APU_WRB1:
      chan = (address & 4) ? 1 : 0;
      square[chan].regs[1] = value;
      break;

   case APU_WRA2:
   case APU_WRB2:
      chan = (address & 4) ? 1 : 0;
      square[chan].regs[2] = value;
      if (square[chan].enabled)
         square[chan].freq = ((((square[chan].regs[3] & 7) << 8) + value) + 1) << 16;
      break;

   case APU_WRA3:
   case APU_WRB3:
      chan = (address & 4) ? 1 : 0;
      square[chan].regs[3] = value;

      if (square[chan].enabled)
      {
         square[chan].vbl_length = vbl_length[value >> 3] * apu_buf_size;
         square[chan].env_vol = 0;
         square[chan].freq = ((((value & 7) << 8) + square[chan].regs[2]) + 1) << 16;
      }

      break;

   /* triangle */
   case APU_WRC0:
      triangle.regs[0] = value;

      if (triangle.enabled)
      {                                          /* ??? */
         if (FALSE == triangle.counter_started)// && triangle.vbl_length)
            triangle.linear_length = ((value & 0x7F) * apu_buf_size) >> 2;
      }

      break;

   case 0x4009:
      /* unused */
      triangle.regs[1] = value;
      break;

   case APU_WRC2:
      triangle.regs[2] = value;
      break;

   case APU_WRC3:
      triangle.regs[3] = value;

      /* this is somewhat of a hack.  there is some latency on the Real
      ** Thing between when trireg0 is written to and when the linear
      ** length counter actually begins its countdown.  we want to prevent
      ** the case where the program writes to the freq regs first, then
      ** to reg 0, and the counter accidentally starts running because of
      ** the sound queue's timestamp processing.
      **
      ** set to a few NES sample -- should be sufficient
      **
      **     3 * (1789772.727 / 44100) = ~122 cycles, just around one scanline
      **
      ** should be plenty of time for the 6502 code to do a couple of table
      ** dereferences and load up the other triregs
      */

      triangle.write_latency = 3;

      if (triangle.enabled)
      {
         triangle.counter_started = FALSE;
         triangle.vbl_length = vbl_length[value >> 3] * apu_buf_size;
         triangle.linear_length = ((triangle.regs[0] & 0x7F) * apu_buf_size) >> 2;
      }

      break;

   /* noise */
   case APU_WRD0:
      noise.regs[0] = value;
      break;

   case 0x400D:
      /* unused */
      noise.regs[1] = value;
      break;

   case APU_WRD2:
      noise.regs[2] = value;
      break;

   case APU_WRD3:
      noise.regs[3] = value;

      if (noise.enabled)
      {
         noise.vbl_length = vbl_length[value >> 3] * apu_buf_size;
         noise.env_vol = 0; /* reset envelope */
      }
      break;

   /* DMC */
   case APU_WRE0:
      dpcm.regs[0] = value;
      if (0 == (value & 0x80))
         dpcm.irq_occurred = FALSE;
      break;

   case APU_WRE1: /* 7-bit DAC */
      //dpcm.regs[1] = value - 0x40;
      dpcm.regs[1] = value;
      break;

   case APU_WRE2:
      dpcm.regs[2] = value;
      //apu_dpcmreset(&dpcm);
      break;

   case APU_WRE3:
      dpcm.regs[3] = value;
      //apu_dpcmreset(&dpcm);
      break;

   /* TODO: clean this up */
   case APU_SMASK:
      apu_enable = value;

      if (value & 0x01)
         square[0].enabled = TRUE;
      else
      {
         square[0].enabled = FALSE;
         square[0].vbl_length = 0;
      }

      if (value & 0x02)
         square[1].enabled = TRUE;
      else
      {
         square[1].enabled = FALSE;
         square[1].vbl_length = 0;
      }

      if (value & 0x04)
         triangle.enabled = TRUE;
      else
      {
         triangle.enabled = FALSE;
         triangle.vbl_length = 0;
         triangle.linear_length = 0;
         triangle.counter_started = FALSE;
         triangle.write_latency = 0;
      }
      if (value & 0x08)
         noise.enabled = TRUE;
      else
      {
         noise.enabled = FALSE;
         noise.vbl_length = 0;
      }

      if (value & 0x10)
      {
         /* only reset dpcm values if DMA is finished */
         if (FALSE == dpcm.enabled)
         {
            dpcm.enabled = TRUE;
            apu_dpcmreset(&dpcm);
         }
      }
      else
         dpcm.enabled = FALSE;

      dpcm.irq_occurred = FALSE;

      break;
   default:
#ifdef NOFRENDO_DEBUG
      log_printf("invalid apu write: $%02X at $%04X\n", value, address);
#endif
      break;
   }
}

void apu_getpcmdata(uint8 **data, int *length)
{
   *data = pcm_data;
   *length = pcm_length;
}

void apu_process8(uint8 *buffer, int length)
{
   static int elapsed_cycles = 0;
   static int prev_sample = 0;
   int next_sample, accum;
   apuqueue_t q;

   /* Set global buffer location */
   pcm_data = buffer;
   pcm_length = length;

   while (length--)
   {
      while (FALSE == apu_isqempty() && (apu_q[q_tail].timestamp <= elapsed_cycles))
      {
         q = apu_dequeue();
         apu_regwrite(q.address, q.value);
      }

      elapsed_cycles += (uint32) apu_incsize;

      accum = 0;
      if (mix_enable[0]) accum += apu_square(&square[0]);
      if (mix_enable[1]) accum += apu_square(&square[1]);
      if (mix_enable[2]) accum += apu_triangle(&triangle);
      if (mix_enable[3]) accum += apu_noise(&noise);
      if (mix_enable[4]) accum += apu_dpcm(&dpcm);
      if (mix_enable[5] && apu_ext) accum += apu_ext();

      /* slap any dsp we need in there */
      if (FALSE != apu_dsp)
      {
         next_sample = accum;
         accum += prev_sample;
         accum >>= 1;

         if (APU_DSP_SIMPLE == apu_dsp)
            prev_sample = next_sample;
         else /* (APU_DSP_MOVING == apu_dsp) */
            prev_sample = accum;
      }

      /* amplify it a tad */
      accum <<= 1;

      /* 8-bit clamps */
      if (accum > 127)
         accum = 127;
      else if (accum < -128)
         accum = -128;

      *buffer++ = accum ^ 0x80;
   }

   /* resync cycle counters */
   elapsed_cycles = cpu_getcycles(FALSE);
}

void apu_process16(uint16 *buffer, int length)
{
   static int elapsed_cycles = 0;
   static int prev_sample = 0;
   int next_sample, accum;
   apuqueue_t q;

   /* set visualization parameters */
   //gui_setvis(&buffer, length);

   while (length--)
   {
      while (FALSE == apu_isqempty() && (apu_q[q_tail].timestamp <= elapsed_cycles))
      {
         q = apu_dequeue();
         apu_regwrite(q.address, q.value);
      }

      elapsed_cycles += (uint32) apu_incsize;

      accum = 0;
      if (mix_enable[0]) accum += apu_square(&square[0]);
      if (mix_enable[1]) accum += apu_square(&square[1]);
      if (mix_enable[2]) accum += apu_triangle(&triangle);
      if (mix_enable[3]) accum += apu_noise(&noise);
      if (mix_enable[4]) accum += apu_dpcm(&dpcm);
      if (mix_enable[5] && apu_ext) accum += apu_ext();

      /* amplify it a tad */
      accum <<= 1;

      if (accum > 127)
         accum = 127;
      else if (accum < -128)
         accum = -128;

      if (FALSE != apu_dsp)
      {
         next_sample = accum;
         accum += prev_sample;
         accum >>= 1;

         if (APU_DSP_SIMPLE == apu_dsp)
            prev_sample = next_sample;
         else /* (APU_DSP_MOVING == apu_dsp) */
            prev_sample = accum;
      }

      *buffer++ = (uint16) (accum ^ 0x80) << 8;
   }

   /* resync cycle counters */
   elapsed_cycles = cpu_getcycles(FALSE);
}

/* Read from $4000-$4017 */
uint8 apu_read(uint32 address)
{
   uint8 value;

   switch (address)
   {
   /* TODO: clean this up */
   case APU_SMASK:
      /* Return 1 in 0-5 bit pos if a channel is playing */
      value = 0;
      if (square[0].enabled && square[0].vbl_length)
         value |= 0x01;
      if (square[1].enabled && square[1].vbl_length)
         value |= 0x02;
      if (triangle.enabled && triangle.vbl_length)
         value |= 0x04;
      if (noise.enabled && noise.vbl_length)
         value |= 0x08;
      //if (dpcm.enabled && dpcm.bits_left)
      //if (dpcm.enabled)
      if (dpcm.enabled && dpcm.length) 
         value |= 0x10;
      /* looks like bit 6 denotes whether or not we're in vblank or not */
      /* for now, we'll just return 1 */
      value |= 0x40;
      if (dpcm.irq_occurred)
         value |= 0x80;

      return value;

   default:
#ifdef NOFRENDO_DEBUG
      log_printf("Invalid register read: $%04X", address);
#endif
      return APU_INVALID;
   }
}

void apu_write(uint32 address, uint8 value)
{
   apuqueue_t q;

   switch (address)
   {
   case 0x4000: case 0x4001: case 0x4002: case 0x4003:
   case 0x4004: case 0x4005: case 0x4006: case 0x4007:
   case 0x4008: case 0x4009: case 0x400A: case 0x400B:
   case 0x400C: case 0x400D: case 0x400E: case 0x400F:
   case 0x4010: case 0x4011: case 0x4012: case 0x4013:
   case 0x4015:
      q.timestamp = cpu_getcycles(FALSE);
      q.address = address;
      q.value = value;
      apu_enqueue(q);
      break;

   default:
      break;
   }
}

static void create_noise(uint8 *buf, const int bits, int size)
{
   static int m = 0x0011;
   int xor_val, i;

   for (i = 0; i < size; i++)
   {
      xor_val = m & 1;
      m >>= 1;
      xor_val ^= (m & 1);
      m |= xor_val << (bits - 1);
#if 0
      if (m & (1 << (bits - 1))) /* if MSB = 1 */
         buf[i] = (uint8) (-m);
      else
         buf[i] = (uint8) m;
#endif
      buf[i] = m;
   }
}

/* set the dsp type */
void apu_setdsp(int dsp_type)
{
   apu_dsp = dsp_type;
}

void apu_reset(void)
{
   memset(&apu_q, 0, APUQUEUE_SIZE * sizeof(apuqueue_t));
   q_head = 0;
   q_tail = 0;

   memset(&square, 0, 2 * sizeof(square_t));
   memset(&triangle, 0, sizeof(triangle_t));
   memset(&noise, 0, sizeof(noise_t));
   memset(&dpcm, 0, sizeof(dpcm_t));

   /* TODO: apu_extreset() */
}

/* Initializes hardware, creates waveforms/voices */
void apu_init(int sample_frequency, int bps, int nch)
{
   int pos;

   apu_buf_size = sample_frequency / REFRESH_RATE;
   apu_incsize = (float) (NES_APUBASE / (float) sample_frequency);

   /* 13-bit noise generator? */
   create_noise(noise_lut, 13, NOISE_LONG);

   for (pos = 0; pos < 6; pos++)
      mix_enable[pos] = TRUE;

   apu_reset();

   apu_ext = NULL;

   apu_setdsp(APU_DSP_NONE);
}

void apu_shutdown(void)
{
}

void apu_setext(apusound_t ext_sound)
{
   apu_ext = ext_sound;
}

/* this exists for external mixing routines */
float apu_getphase(void)
{
   return apu_incsize;
}


