/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** nsf_play.c
**
** NES hardware emulation for NSF routines
*/

#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "cpu.h"
#include "log.h"

#include "nsf_play.h"
#include "nes_apu.h"
#include "vrcvisnd.h"

/* TODO: build cpu read/write handler structs dynamically */
static cpu_context *cpu;
static nsf_t *nsf_info;

static __inline__ uint8 read_ram(uint32 address)
{
   return cpu->mem_page[0][address & 0x7FF];
}

static __inline__ void write_ram(uint32 address, uint8 value)
{
   cpu->mem_page[0][address & 0x7FF] = value;
}

/* TODO: memcpy?  are you out of your fucking mind?
**       you do know how to bankswitch, right?  RIGHT?!?!
*/
static __inline__ void nsf_bankswitch(uint32 address, uint8 value)
{
   uint32 bank_src, bank_dest, bank_offset;
   int bank_length;

   if (FALSE == nsf_info->bankswitched)
      return;

   /* destination in nes address space */
   bank_dest = 0x8000 + ((address & 7) << 12);

   /* offset of banks in nsf data */
   bank_offset = nsf_info->load_addr & 0x0FFF;

   if (0 == value) /* special case for bank 0 */
   {
      ASSERT(nsf_info->load_addr < 0x9000);
      bank_src = 0;
      bank_length = 0x1000 - bank_offset;
      bank_dest += bank_offset;
   }
   else
   {
      bank_src = (value << 12) - bank_offset;
      ASSERT((int) (nsf_info->length - bank_src) > 0);

      if (nsf_info->length - bank_src < 0x1000)
         bank_length = nsf_info->length - bank_src;
      else
         bank_length = 0x1000;
   }

/*
   log_printf("banking $%04X bytes at nes:$%04X from nsf:bank %d offset $%04X\n",
              bank_length, bank_dest, value, bank_src);
*/
   memcpy(cpu->mem_page[0] + bank_dest, nsf_info->data + bank_src, bank_length);
}

static __inline__ void ext_write(uint32 address, uint8 value)
{
   switch (nsf_info->ext_sound_type)
   {
   case EXT_SOUND_VRCVI:
      vrcvi_soundwrite(address, value);
      break;
   case EXT_SOUND_VRCVII:
      break;
   case EXT_SOUND_FDS:
      break;
   case EXT_SOUND_NONE:
   default:
      break;
   }
}

static cpu_readhandler nsf_readhandler[] =
{
   /* $0 - $7FF is RAM */
   { 0x0800, 0x1FFF, read_ram },
   { 0x4000, 0x4017, apu_read },
   { -1,     -1,     NULL }
};

static cpu_writehandler nsf_writehandler[] =
{
   /* $0 - $7FF is RAM */
   { 0x0800, 0x1FFF, write_ram },
   { 0x4000, 0x4017, apu_write },
   { 0x4018, 0x40FF, ext_write }, /* FDS sound registers ! */
   { 0x5FF8, 0x5FFF, nsf_bankswitch },
   { 0x8000, 0xFFFF, ext_write },
   { -1,     -1,     NULL}
};

/* sets up a simple loop that calls the desired routine and spins */
void nsf_setup(uint32 address, uint8 accum)
{
   cpu_getcontext(cpu);

   cpu->mem_page[0][0x4100] = 0x20; /* jsr */
   cpu->mem_page[0][0x4101] = address & 0xFF;
   cpu->mem_page[0][0x4102] = address >> 8;   /* init addr */
   cpu->mem_page[0][0x4103] = 0xFF; /* kill switch for CPU emulation */

   cpu->pc_reg = 0x4100;
   cpu->a_reg = accum;

   cpu_setcontext(cpu);
}

static void nes_zero(void)
{
   uint32 address;

   memset(cpu->mem_page[0], 0, 0x800);

   /* Set up sound regs */
   for (address = 0x4000; address <= 0x4013; address++)
      apu_write(address, 0);

   apu_write(0x4015, 0x0F);
}

void nsf_reset(void)
{
   nes_zero();

   switch (nsf_info->ext_sound_type)
   {
   case EXT_SOUND_VRCVI:
      vrcvi_reset();
      break;
   case EXT_SOUND_VRCVII:
      break;
   case EXT_SOUND_FDS:
      break;
   case EXT_SOUND_NONE:
   default:
      break;
   }

   if (nsf_info->bankswitched)
   {
      nsf_bankswitch(0x5FF8, nsf_info->bankswitch_info[0]);
      nsf_bankswitch(0x5FF9, nsf_info->bankswitch_info[1]);
      nsf_bankswitch(0x5FFA, nsf_info->bankswitch_info[2]);
      nsf_bankswitch(0x5FFB, nsf_info->bankswitch_info[3]);
      nsf_bankswitch(0x5FFC, nsf_info->bankswitch_info[4]);
      nsf_bankswitch(0x5FFD, nsf_info->bankswitch_info[5]);
      nsf_bankswitch(0x5FFE, nsf_info->bankswitch_info[6]);
      nsf_bankswitch(0x5FFF, nsf_info->bankswitch_info[7]);
   }
   else
   {
      ASSERT(nsf_info->load_addr + nsf_info->length <= 0x10000);
      memcpy(cpu->mem_page[0] + nsf_info->load_addr, nsf_info->data, nsf_info->length);
   }

   nsf_setup(nsf_info->init_addr, (uint8) (nsf_info->current_song - 1));

   /* execute 1 frame or so (let routine do setup) */
   cpu_execute((int) NES_FRAME_CYCLES);
}


int nes_frame(void)
{
   nsf_setup(nsf_info->play_addr, 0); /* Don't care about accum */

   /* 1 Frame */
   /*
   log_printf("asked for %d cycles, got %d\n", (uint32) NES_FRAME_CYCLES,
      cpu_execute((int) NES_FRAME_CYCLES));
   */
   cpu_execute((int) NES_FRAME_CYCLES);

   return 0;
}

/* Reset NES hardware */
void nes_reset(int reset_type)
{
   cpu_reset();
   apu_reset();

   nsf_reset();
}

/* Deallocate memory */
void nes_shutdown(void)
{
   if (cpu->mem_page[0])
      free(cpu->mem_page[0]);
   free(cpu);
   cpu = NULL;
}

/* Initialize NES CPU, hardware, etc. */
int nes_init(nsf_t *nsf)
{
   int loop;

   nsf_info = nsf;

   cpu_init();

   cpu = (cpu_context *) malloc(cpu_contextsize());
   if (NULL == cpu)
      return -1;
   memset(cpu, 0, cpu_contextsize());

   /* aw, hell -- just go ahead and allocate 64kB anyway... */
   cpu->mem_page[0] = (uint8 *) malloc(CPU_RAMSIZE);
   if (NULL == cpu->mem_page[0])
   {
      nes_shutdown();
      return -1;
   }

   cpu->read_handler = nsf_readhandler;
   cpu->write_handler = nsf_writehandler;

#ifdef NES6502_DEBUG
   cpu->trace = TRUE;
#endif /* NES6502_DEBUG */

   for (loop = 1; loop < 8; loop++)
   {
#ifdef NES6502_FASTBANKS
      cpu->mem_page[loop] = cpu->mem_page[0];
#else
      cpu->mem_page[loop] = cpu->mem_page[0] + (0x2000 * loop);
#endif /* NES6502_FASTBANKS */
   }

   cpu_setcontext(cpu);

   return 0;
}
