/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
** 
** vrcvisnd.c
**
** VRCVI sound hardware emulation
*/

#include "types.h"
#include "vrcvisnd.h"
#include "nes_apu.h"


/* output channel structures */
static vrcvisquare_t square[2];
static vrcvisawtooth_t sawtooth;
static float vrcvi_incsize;

/* VRCVI square wave generation */
static int8 vrcvi_square(vrcvisquare_t *chan)
{
   float freq;
   int8 output;

   /* reg0: 0-3=volume, 4-6=duty cycle
   ** reg1: 8 bits of freq
   ** reg2: 0-3=high freq, 7=enable
   */

   freq = (float) (((chan->reg[2] & 0x0F) << 8) + chan->reg[1] + 1);

   chan->phaseacc -= vrcvi_incsize; /* # of clocks per wave cycle */
   while (chan->phaseacc < 0)
   {
      chan->phaseacc += freq;
      chan->adder = (chan->adder + 1) & 0x0F;
   }

   /* return if not enabled */
   if (0 == (chan->reg[2] & 0x80))
      return 0;

   output = chan->reg[0] & 0x0F;

   if (chan->adder < ((chan->reg[0] >> 4) + 1))
      output = -output;

   return output;
}



/* VRCVI sawtooth wave generation */
static int8 vrcvi_sawtooth(vrcvisawtooth_t *chan)
{
   float freq;

   /* reg0: 0-5=phase accumulator bits
   ** reg1: 8 bits of freq
   ** reg2: 0-3=high freq, 7=enable
   */

   freq = (float) (((chan->reg[2] & 0x0F) << 8) + chan->reg[1] + 1);

   chan->phaseacc -= vrcvi_incsize; /* # of clocks per wav cycle */
   while (chan->phaseacc < 0)
   {
/*      chan->phaseacc += freq << 1; */
      chan->phaseacc += freq;
      chan->phaseacc += freq;
      chan->output_acc += (chan->reg[0] & 0x3F);
      
      if (7 == ++chan->adder)
      {
         chan->adder = 0;
         chan->output_acc = 0;
      }
   }

   /* return if not enabled */
   if (0 == (chan->reg[2] & 0x80))
      return 0;
   else
      return (int8) ((chan->output_acc >> 3) & 0x1F);
}

/* mix vrcvi sound channels together */
int8 vrcvi_soundprocess(void)
{
   int8 output;

   output = vrcvi_square(&square[0]);
   output += vrcvi_square(&square[1]);
   output += vrcvi_sawtooth(&sawtooth);

   return output;
}

/* reset state of vrcvi sound channels */
void vrcvi_reset(void)
{
   memset(&square[0], 0, sizeof(vrcvisquare_t));
   memset(&square[1], 0, sizeof(vrcvisquare_t));
   memset(&sawtooth, 0, sizeof(vrcvisawtooth_t));

   /* kind of a hack, but gets rid of high pitched squealing */
   square[0].reg[1] = 0xFF;
   square[1].reg[1] = 0xFF;
   sawtooth.reg[1] = 0xFF;

   /* get the phase period from the apu */
   vrcvi_incsize = apu_getphase();
}

/* write to registers */
void vrcvi_soundwrite(uint32 address, uint8 value)
{
   switch (address & 0xB000)
   {
   case 0x9000:
      square[0].reg[address & 3] = value;
      break;
   case 0xA000:
      square[1].reg[address & 3] = value;
      break;
   case 0xB000:
      sawtooth.reg[address & 3] = value;
      break;
   default:
      break;
   }
}
